/* 
 * tclVmsFile.c --
 *
 *      This file contains wrappers around UNIX file handling functions.
 *      These wrappers mask differences between Windows and UNIX.
 *
 */


#include "tclInt.h"
#include "tclPort.h"
#include <unixlib.h>
#include <unixio.h>

#include <ctype.h>
#include <string.h>
#include <starlet.h>
#include <rms.h>

/*
 * The variable below caches the name of the current working directory
 * in order to avoid repeated calls to getcwd.  The string is malloc-ed.
 * NULL means the cache needs to be refreshed.
 */

static char *currentDir =  NULL;
static int currentDirExitHandlerSet = 0;

/*
 * The variable below is set if the exit routine for deleting the string
 * containing the executable name has been registered.
 */

static int executableNameExitHandlerSet = 0;

extern pid_t waitpid _ANSI_ARGS_((pid_t pid, int *stat_loc, int options));

/*
 * Static routines for this file:
 */

static void	FreeCurrentDir _ANSI_ARGS_((ClientData clientData));
static void	FreeExecutableName _ANSI_ARGS_((ClientData clientData));
static void	toLowerCase _ANSI_ARGS_((char *str));


/*
 *----------------------------------------------------------------------
 *
 * Tcl_WaitPid --
 *
 *	Implements the waitpid system call on Unix systems.
 *
 * Results:
 *	Result of calling waitpid.
 *
 * Side effects:
 *	Waits for a process to terminate.
 *
 *----------------------------------------------------------------------
 */

int
Tcl_WaitPid(pid, statPtr, options)
    int pid;
    int *statPtr;
    int options;
{
#if 0
    int result;
    pid_t real_pid;

    real_pid = (pid_t) pid;
    while (1) {
	result = (int) waitpid(real_pid, statPtr, options);
	if ((result != -1) || (errno != EINTR)) {
	    return result;
	}
    }
#else
    return -1;
#endif
}

/*
 *----------------------------------------------------------------------
 *
 * FreeCurrentDir --
 *
 *	Frees the string stored in the currentDir variable. This routine
 *	is registered as an exit handler and will be called during shutdown.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Frees the memory occuppied by the currentDir value.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static void
FreeCurrentDir(clientData)
    ClientData clientData;	/* Not used. */
{
    if (currentDir != (char *) NULL) {
        ckfree(currentDir);
        currentDir = (char *) NULL;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * FreeExecutableName --
 *
 *	Frees the string stored in the tclExecutableName variable. This
 *	routine is registered as an exit handler and will be called
 *	during shutdown.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Frees the memory occuppied by the tclExecutableName value.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static void
FreeExecutableName(clientData)
    ClientData clientData;	/* Not used. */
{
    if (tclExecutableName != (char *) NULL) {
        ckfree(tclExecutableName);
        tclExecutableName = (char *) NULL;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TclChdir --
 *
 *	Change the current working directory.
 *
 * Results:
 *	The result is a standard Tcl result.  If an error occurs and 
 *	interp isn't NULL, an error message is left in interp->result.
 *
 * Side effects:
 *	The working directory for this application is changed.  Also
 *	the cache maintained used by TclGetCwd is deallocated and
 *	set to NULL.
 *
 *----------------------------------------------------------------------
 */

int
TclChdir(interp, dirName)
    Tcl_Interp *interp;		/* If non NULL, used for error reporting. */
    char *dirName;     		/* Path to new working directory. */
{
    char *e, *r, *u, uName[512];

    if (currentDir != NULL) {
	ckfree(currentDir);
	currentDir = NULL;
    }

    /*
     * Need to convert disk:[dir]file.dir to disk:[dir.file]
     *
     * Convert to Unix format and wing it from there
     */
    if (u = vmsNameToUnix(dirName)) {

    	/*
    	 * We strip out the extension if it's there.
    	 */
    	strcpy(uName, u);
    	u = uName;
	if (r = strrchr(u, '/'))
	    r++;
	else
	    r = u;
	if (e = strchr(r, '.'))
	    *e = 0;
    }
    else
    	u = dirName;
    if (chdir(u) != 0) {
	if (interp != NULL) {
	    Tcl_AppendResult(interp, "couldn't change working directory to \"",
		    dirName, "\": ", Tcl_PosixError(interp), (char *) NULL);
	}
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TclGetCwd --
 *
 *	Return the path name of the current working directory.
 *
 * Results:
 *	The result is the full path name of the current working
 *	directory, or NULL if an error occurred while figuring it out.
 *	The returned string is owned by the TclGetCwd routine and must
 *	not be freed by the caller.  If an error occurs and interp
 *	isn't NULL, an error message is left in interp->result.
 *
 * Side effects:
 *	The path name is cached to avoid having to recompute it
 *	on future calls;  if it is already cached, the cached
 *	value is returned.
 *
 *----------------------------------------------------------------------
 */

char *
TclGetCwd(interp)
    Tcl_Interp *interp;		/* If non NULL, used for error reporting. */
{
    char buffer[MAXPATHLEN+1];

    if (currentDir == NULL) {
        if (!currentDirExitHandlerSet) {
            currentDirExitHandlerSet = 1;
            Tcl_CreateExitHandler(FreeCurrentDir, (ClientData) NULL);
        }
	if (getcwd(buffer, MAXPATHLEN+1, 0) == NULL) {
	    if (interp != NULL) {
		if (errno == ERANGE) {
		    interp->result = "working directory name is too long";
		} else {
		    Tcl_AppendResult(interp,
			    "error getting working directory name: ",
			    Tcl_PosixError(interp), (char *) NULL);
		}
	    }
	    return NULL;
	}
	currentDir = (char *) ckalloc((unsigned) (strlen(buffer) + 1));
	strcpy(currentDir, buffer);
	toLowerCase(currentDir);
    }
    return currentDir;
}

/*
 *----------------------------------------------------------------------
 *
 * TclOpenFile --
 *
 *	Implements a mechanism to open files on Unix systems.
 *
 * Results:
 *	The opened file.
 *
 * Side effects:
 *	May cause a file to be created on the file system.
 *
 *----------------------------------------------------------------------
 */

Tcl_File
TclOpenFile(fname, mode)
    char *fname;			/* The name of the file to open. */
    int mode;				/* In what mode to open the file? */
{
    int fd;

    fd = open(fname, mode, 0600);
    if (fd != -1) {
#ifndef VMS
	fcntl(fd, F_SETFD, FD_CLOEXEC);
#endif	/* !VMS */
	return Tcl_GetFile((ClientData)fd, TCL_VMS_FILE);
    }
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * TclCloseFile --
 *
 *	Implements a mechanism to close a UNIX file.
 *
 * Results:
 *	Returns 0 on success, or -1 on error, setting errno.
 *
 * Side effects:
 *	The file is closed.
 *
 *----------------------------------------------------------------------
 */

int
TclCloseFile(file)
    Tcl_File file;	/* The file to close. */
{
    int type;
    int fd;
    int result;

    fd = (int) Tcl_GetFileInfo(file, &type);
    if (type != TCL_VMS_FILE) {
	panic("Tcl_CloseFile: unexpected file type");
    }

    /*
     * Refuse to close the fds for stdin, stdout and stderr.
     */
    
    if ((fd == 0) || (fd == 1) || (fd == 2)) {
        return 0;
    }
    
    result = close(fd);
    Tcl_DeleteFileHandler(file);
    Tcl_FreeFile(file);
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * TclReadFile --
 *
 *	Implements a mechanism to read from files on Unix systems. Also
 *	simulates blocking behavior on non-blocking files when asked to.
 *
 * Results:
 *	The number of characters read from the specified file.
 *
 * Side effects:
 *	May consume characters from the file.
 *
 *----------------------------------------------------------------------
 */
	/* ARGSUSED */
int
TclReadFile(file, shouldBlock, buf, toRead)
    Tcl_File file;	/* The file to read from. */
    int shouldBlock;		/* Not used. */
    char *buf;			/* The buffer to store input in. */
    int toRead;			/* Number of characters to read. */
{
    int type, fd, result;

    fd = (int) Tcl_GetFileInfo(file, &type);
    if (type != TCL_VMS_FILE) {
	panic("Tcl_ReadFile: unexpected file type");
    }

    return read(fd, buf, (size_t) toRead);
}

/*
 *----------------------------------------------------------------------
 *
 * TclWriteFile --
 *
 *	Implements a mechanism to write to files on Unix systems. Also
 *	simulates blocking behavior on non-blocking files when asked to.
 *
 * Results:
 *	The number of characters written to the specified file.
 *
 * Side effects:
 *	May produce characters on the file.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
TclWriteFile(file, shouldBlock, buf, toWrite)
    Tcl_File file;	/* The file to write to. */
    int shouldBlock;		/* Not used. */
    char *buf;			/* Where output is stored. */
    int toWrite;		/* Number of characters to write. */
{
    int type, fd;

    fd = (int) Tcl_GetFileInfo(file, &type);
    if (type != TCL_VMS_FILE) {
	panic("Tcl_WriteFile: unexpected file type");
    }

    return write(fd, buf, (size_t) toWrite);
}

/*
 *----------------------------------------------------------------------
 *
 * TclSeekFile --
 *
 *	Sets the file pointer on the indicated UNIX file.
 *
 * Results:
 *	The new position at which the file will be accessed, or -1 on
 *	failure.
 *
 * Side effects:
 *	May change the position at which subsequent operations access the
 *	file designated by the file.
 *
 *----------------------------------------------------------------------
 */

int
TclSeekFile(file, offset, whence)
    Tcl_File file;	/* The file to seek on. */
    int offset;			/* How far to seek? */
    int whence;			/* And from where to seek? */
{
    int type, fd;

    fd = (int) Tcl_GetFileInfo(file, &type);
    if (type != TCL_VMS_FILE) {
	panic("Tcl_SeekFile: unexpected file type");
    }

    return lseek(fd, offset, whence);
}

/*
 *----------------------------------------------------------------------
 *
 * TclCreateTempFile --
 *
 *	This function creates a temporary file initialized with an
 *	optional string, and returns a file handle with the file pointer
 *	at the beginning of the file.
 *
 * Results:
 *	A handle to a file.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

Tcl_File
TclCreateTempFile(contents)
    char *contents;		/* String to write into temp file, or NULL. */
{
    char fileName[L_tmpnam];
    Tcl_File file;
    size_t length = (contents == NULL) ? 0 : strlen(contents);

    tmpnam(fileName);
    file = TclOpenFile(fileName, O_RDWR|O_CREAT|O_TRUNC);
    unlink(fileName);

    if ((file != NULL) && (length > 0)) {
	int fd = (int)Tcl_GetFileInfo(file, NULL);
	while (1) {
	    if (write(fd, contents, length) != -1) {
		break;
	    } else if (errno != EINTR) {
		close(fd);
		Tcl_FreeFile(file);
		return NULL;
	    }
	}
	lseek(fd, 0, SEEK_SET);
    }
    return file;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_FindExecutable --
 *
 *	This procedure computes the absolute path name of the current
 *	application, given its argv[0] value.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The variable tclExecutableName gets filled in with the file
 *	name for the application, if we figured it out.  If we couldn't
 *	figure it out, Tcl_FindExecutable is set to NULL.
 *
 *----------------------------------------------------------------------
 */

void
Tcl_FindExecutable(argv0)
    char *argv0;		/* The value of the application's argv[0]. */
{
    char *name, *p, *cwd;
    int length;

    if (tclExecutableName != NULL) {
	ckfree(tclExecutableName);
	tclExecutableName = NULL;
    }

    tclExecutableName = (char *) ckalloc((unsigned) (strlen(argv0) + 1));
    strcpy(tclExecutableName, argv0);

    if (!executableNameExitHandlerSet) {
        executableNameExitHandlerSet = 1;
        Tcl_CreateExitHandler(FreeExecutableName, (ClientData) NULL);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TclGetUserHome --
 *
 *	This function takes the passed in user name and finds the
 *	corresponding home directory specified in the password file.
 *
 * Results:
 *	The result is a pointer to a static string containing
 *	the new name.  If there was an error in processing the
 *	user name then the return value is NULL.  Otherwise the
 *	result is stored in bufferPtr, and the caller must call
 *	Tcl_DStringFree(bufferPtr) to free the result.
 *
 * Side effects:
 *	Information may be left in bufferPtr.
 *
 *----------------------------------------------------------------------
 */

char *
TclGetUserHome(name, bufferPtr)
    char *name;			/* User name to use to find home directory. */
    Tcl_DString *bufferPtr;	/* May be used to hold result.  Must not hold
				 * anything at the time of the call, and need
				 * not even be initialized. */
{
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * TclMatchFiles --
 *
 *	This routine is used by the globbing code to search a
 *	directory for all files which match a given pattern.
 *
 * Results: 
 *	If the tail argument is NULL, then the matching files are
 *	added to the interp->result.  Otherwise, TclDoGlob is called
 *	recursively for each matching subdirectory.  The return value
 *	is a standard Tcl result indicating whether an error occurred
 *	in globbing.
 *
 * Side effects:
 *	None.
 *
 * To be done:
 *	o tail arg is ignored
 *
 *---------------------------------------------------------------------- */

int
TclMatchFiles(interp, separators, dirPtr, pattern, tail)
    Tcl_Interp *interp;		/* Interpreter to receive results. */
    char *separators;		/* Path separators to pass to TclDoGlob. */
    Tcl_DString *dirPtr;	/* Contains path to directory to search. */
    char *pattern;		/* Pattern to match against. */
    char *tail;			/* Pointer to end of pattern. */
{
    char *d;
    char *dirName;
    char *mask = pattern;
    char resultantString[NAM$C_MAXRSS];
    char expandedString[NAM$C_MAXRSS];
    int status;
    int stripDir;
    struct FAB fab;
    struct NAM nam;

    if (dirPtr->string[0] == '\0') {
	dirName = "sys$disk:[]";
	stripDir = 1;
    }
    else {
	dirName = dirPtr->string;
	stripDir = 0;
    }

    fab = cc$rms_fab;
    nam = cc$rms_nam;

    if (strcmp(mask, "*") == 0)
    	mask = "*.*";
    fab.fab$l_fop = FAB$M_NAM;
    fab.fab$l_nam = &nam;
    fab.fab$l_fna = mask;
    fab.fab$b_fns = strlen(mask);

    nam.nam$b_rss = NAM$C_MAXRSS;
    nam.nam$l_rsa = resultantString;
    nam.nam$b_ess = NAM$C_MAXRSS;
    nam.nam$l_esa = expandedString;

    status = sys$parse(&fab);
    if (status != RMS$_NORMAL) {
	Tcl_AppendResult(interp, "couldn't read directory \"",
		dirPtr->string, "\": ", Tcl_PosixError(interp), (char *) NULL);
	return TCL_ERROR;
    }
    
    while ((status = sys$search(&fab)) == RMS$_NORMAL) {
	resultantString[nam.nam$b_rsl] = '\0';
#if 0
	if (s = strchr(resultantString, ';'))
	    *s = '.';
#endif
	toLowerCase(resultantString);		/* Stop the shouting */
	if (stripDir) {
	    if (d = strchr(resultantString, ']'))
		d++;
	    else if (d = strchr(resultantString, ':'))
		d++;
	    else
		d = resultantString;
	}
	else
	    d = resultantString;
	Tcl_AppendElement(interp, d);
    }
    status = sys$close(&fab);
    return TCL_OK;
}

static void
toLowerCase(char *str)
{
    char c;
    while (c = *str) {
	if (isupper(c))
	    *str = tolower(c);
	str++;
    }
}

void
Tcl_VmsTranslateFileName(Tcl_Interp *interp, char *name, Tcl_DString *bufferPtr)
{
    /*
     *	Convert a filename to Unix format
     */
    char vmsName[L_tmpnam];
    char unixName[L_tmpnam];

    /*
     *	Convert from Unix to VMS to normalize hybrid filenames.
     */
    if (strchr(name, '/')) {
	if (Tcl_UnixToVms(name, vmsName)) {
	}
	else {
	}
    }
    else
	strcpy(vmsName, name);
    
    /*
     *	Name is now is in VMS format, convert it back to Unix and return
     */
    Tcl_VmsToUnix(vmsName, unixName);

    Tcl_DStringInit(bufferPtr);
    Tcl_DStringAppend(bufferPtr, unixName, -1);
}
